try:
    import pandas as pd
    from plotly import graph_objects as go
    from plotly.subplots import make_subplots
    import plotly.express as px
    import plotly.io as pio
except ImportError:
    pd = None
    go = None
    make_subplots = None
    px = None
    pio = None


def _validate_libraries():
    if pd is None:
        raise ImportError("pandas is not installed. Please install pandas to use this function.")
    if None in [go, make_subplots, px, pio]:
        raise ImportError("plotly is not installed. Please install plotly to use this function.")


class Color:
    VERY_GREEN = "#26A69A"
    GREEN = "#B2DFDB"
    VERY_RED = "#EF5350"
    RED = "#FFCDD2"
    BG_YELLOW = "#FFFDEE"
    YELLOW_1 = "#f6c309"
    YELLOW_2 = "#fb9800"
    YELLOW_3 = "#fb6500"
    YELLOW_4 = "#f60c0c"


def line(
    series: pd.Series,
    index: pd.Series = None,
    name: str = None,
    width: float = 1,
    color: str = None,
    dash: str = None,
    shape: str = None,
    gl: bool = True,
    **kwargs,
):
    """Helper function to create line plots from a pandas series."""
    _validate_libraries()
    if index is not None:
        assert len(series) == len(index), "Series length should be equal to index length"

    line_dict = dict(width=width)
    if color is not None:
        line_dict["color"] = color
    if dash:
        line_dict["dash"] = dash
    if shape:
        line_dict["shape"] = shape
    kwargs = kwargs | dict(
        x=index if index is not None else series.index,
        y=series,
        mode=kwargs.get("mode", "lines"),
        name=name or series.name,
        line=line_dict if line_dict else None,
    )

    return go.Scattergl(**kwargs) if gl else go.Scatter(**kwargs)


def scatter(
    series: pd.Series,
    index: pd.Series = None,
    name: str = None,
    color: str = None,
    size: int = 10,
    symbol: str = "circle",
    gl: bool = True,
    **kwargs,
):
    """Helper function to create scatter plots from a pandas series."""
    _validate_libraries()
    dikt = dict(
        x=index if index is not None else series.index,
        y=series,
        mode="markers",
        name=name or series.name,
        marker_symbol=symbol,
        marker_size=size,
        marker_color=color,
    )
    kwargs = dikt | kwargs
    return go.Scattergl(**kwargs) if gl else go.Scatter(**kwargs)


def subplot(
    columns: int = 1,
    rows: int = 1,
    secondary: bool = True,
    shared_xaxes: bool = None,
    vertical_spacing: float = None,
    **kwargs,
):
    """Helper function to create subplots with secondary y-axis support."""
    _validate_libraries()
    return make_subplots(
        cols=columns,
        rows=rows,
        specs=[[{"secondary_y": secondary} for _ in range(columns)] for _ in range(rows)],
        shared_xaxes=shared_xaxes,
        vertical_spacing=vertical_spacing,
        **kwargs,
    )


def insert_title(
    fig,
    title,
    x=0.01,
    y=0.99,
    center: bool = False,
    size: int = 20,
    bgcolor: str = "rgba(255, 255, 255, 0.7)",
    color: str = "black",
    font_family: str = "Courier New, monospace",
):
    """Insert title in the figure."""

    _validate_libraries()
    fig.add_annotation(
        x=x,
        y=y,
        xref="paper",
        yref="paper",
        showarrow=False,
        text=title,
        font=dict(size=size, color=color, family=font_family),
        bgcolor=bgcolor,
        borderpad=2,
    )
    return fig



import asyncio
import datetime as dt
import os
from dataclasses import dataclass, replace
from logging import Logger
from typing import Dict, Callable, Tuple, List, Union, Type, Sequence, Iterable
import time
from functools import wraps
from operator import attrgetter
import random
from zoneinfo import ZoneInfo


MIN_FLOAT = float("-inf")
MAX_FLOAT = float("inf")
TZ = ZoneInfo(os.getenv("TZ")) if os.getenv("TZ") else dt.datetime.now().astimezone().tzinfo
TypeException = Type[Exception]

TaskType = Tuple[Callable, List, Dict]
Comparable = Union[None, dt.date, dt.datetime, int, float]


def camel_or_space_to_snake(s):
    s = s.replace(" ", "_")
    return "".join(["_" + c.lower() if c.isupper() else c for c in s]).lstrip("_")


@dataclass
class Interval:
    start: Comparable
    end: Comparable

    def intersects(self, other: "Interval") -> bool:
        return self.start <= other.end and other.start <= self.end

    def intersection(self, other: "Interval") -> "Interval":
        if not self.intersects(other):
            raise ValueError("No intersection possible")

        return replace(self, start=max(self.start, other.start), end=min(self.end, other.end))

    @classmethod
    def union(cls, intervals: Sequence["Interval"], allow_gaps: bool = False) -> "Interval":
        # TODO: Fix this: should have a signature same as intersection.
        if len(intervals) == 0:
            raise AssertionError("Length of intervals should at least be 1")
        if len(intervals) == 1:
            return intervals[0]

        intervals = sorted(intervals, key=attrgetter("start"))
        if not allow_gaps:
            for i in range(len(intervals) - 1):
                first = intervals[i]
                second = intervals[i + 1]
                if first.end < second.start:
                    raise ValueError(f"Gaps not supported in interval union")

        return cls(start=intervals[0].start, end=intervals[-1].end)

    @property
    def range(self) -> Comparable:
        return self.end - self.start

    def _validate(self):
        assert self.start < self.end, "start value should be less than equal to end"

    def __post_init__(self):
        self._validate()

    def __or__(self, other) -> "Interval":
        return self.__class__.union([self, other])

    def __and__(self, other) -> "Interval":
        return self.intersection(other)

    def __sub__(self, other: Comparable) -> "Interval":
        return self.__class__(start=self.start - other, end=self.end - other)

    def __add__(self, other: Comparable) -> "Interval":
        return self.__class__(start=self.start + other, end=self.end + other)

    def __mul__(self, other) -> "Interval":
        return self.__class__(start=self.start * other, end=self.end * other)

    def __contains__(self, item: Union[Comparable, "Interval"]):
        if isinstance(item, Interval):
            return item.start in self and item.end in self
        return self.start <= item <= self.end

    def __str__(self):
        return f"[{self.start} — {self.end}]"


class retry:  # noqa
    """Retry calling the decorated function using an exponential backoff. Asyncio compatible."""

    def __init__(
        self,
        exception_to_check: Union[TypeException, Tuple[TypeException]],
        tries: int = 4,
        delay: float = 1,
        backoff: float = 2,
        logger: Logger = None,
    ):
        """
        Args:
            exception_to_check: the exception to check. Maybe a tuple of exceptions to check.
            tries: number of times to try (not retry) before giving up
            delay: initial delay between retries in seconds
            backoff: backoff multiplier e.g. value of 2 will double the delay for each retry
            logger: logger to use. If None, print.
        """
        self._tries = tries
        self._delay = delay
        self._backoff = backoff
        self._ignore_exc_types = exception_to_check
        self._log_fn = logger.warning if logger else print

    def __call__(self, fn):
        is_async = asyncio.iscoroutinefunction(fn)
        if not is_async:

            @wraps(fn)
            def decorated(*a, **k):
                last_exception = None
                delay = self._delay
                for attempt in range(self._tries):
                    try:  # noqa
                        return fn(*a, **k)
                    except self._ignore_exc_types as e:
                        last_exception = e
                        msg = f"Error {e.__class__.__name__}: retrying {fn.__name__} - in {delay}s - attempt {attempt + 1}/{self._tries}"
                        self._log_fn(msg)
                    time.sleep(delay)
                    delay *= self._backoff
                if last_exception is not None:
                    raise last_exception

            return decorated

        else:

            @wraps(fn)
            async def decorated(*a, **k):
                last_exception = None
                delay = self._delay
                for attempt in range(self._tries):
                    try:  # noqa
                        return await fn(*a, **k)
                    except self._ignore_exc_types as e:
                        last_exception = e
                        msg = f"{str(e.__class__.__name__)}: retrying {fn.__name__} - in {delay}s - attempt {attempt + 1}/{self._tries}"
                        self._log_fn(msg)
                    await asyncio.sleep(delay)
                    delay *= self._backoff
                if last_exception is not None:
                    raise last_exception

            return decorated


JitterT = Union[float, int, dt.timedelta]


def add_jitter(n: JitterT, jitter_percentage: int = 20) -> JitterT:
    """Add jitter to seconds."""
    return n * (1 + random.randint(-jitter_percentage, jitter_percentage) / 100)


def group(iterable: Iterable, key: Callable) -> Dict:
    """Group elements of an iterable by a key function."""
    groups = {}
    for item in iterable:
        k = key(item)
        if k not in groups:
            groups[k] = []
        groups[k].append(item)
    return groups



import math
from datetime import datetime
from operator import attrgetter
from pathlib import Path
from typing import Optional, Hashable, Union, Tuple

from vayu.common import Interval
from vayu.time import TimeWindow, from_timestamp

try:
    import pandas as pd

    DayOfWeek = pd.CategoricalDtype(
        ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"], ordered=True
    )
except ImportError:
    pd = None
    DayOfWeek = None


def _validate_libraries():
    if pd is None:
        raise ImportError("pandas is not installed. Please install pandas to use this function.")


def concat_frame_from_dir(
    path, prefix: str = None, extension="parquet", progress=False
) -> pd.DataFrame:
    """Concatenate all dataframes in a directory"""

    path = Path(path)
    files = path.glob(f"*.{extension}")
    if prefix is not None:
        files = [f for f in files if f.name.startswith(prefix)]
    files = sorted(files, key=attrgetter("name"))
    reader_func = None
    match extension:
        case "feather":
            reader_func = pd.read_feather
        case "csv":
            reader_func = pd.read_csv
        case "pickle":
            reader_func = pd.read_pickle
        case "parquet":
            reader_func = pd.read_parquet

    frags = []
    n = len(files)
    for i, f in enumerate(files):
        frags.append(reader_func(f))
        if progress:
            print(f"Finished reading [{i + 1}/{n}] {f.name}")

    return pd.concat([f for f in frags if not f.empty], ignore_index=True)


def slice_frame(
    interval: Interval,
    df: "pd.DataFrame",
    level: Optional[int] = None,
    key: Optional[Hashable] = None,
    axis=0,
    exclude: bool = False,
) -> "pd.DataFrame":
    """Slice a dataframe using this interval.

    Args:
        interval: The interval to slice
        df: The dataframe to slice.
        level: The index or column level to slice on
        key: The row index or column to slice on to slice on
        axis: The axis to slice on (0 for index, 1 for columns)
        exclude: If True, exclude the interval instead of including it

    Notes:
        - If key is not specified, the level (or index if level is None) should be sorted in increasing order
        - If neither key nor level is specified, the 0th level of index is used for sliced on.
    """
    _validate_libraries()
    if level is not None:
        slicer = [slice(None)] * df.index.nlevels
        slicer[level] = slice(interval.start, interval.end)
        slicer = pd.IndexSlice[tuple(slicer)]
    elif key is not None:
        key = df[key] if axis == 0 else df.loc[key]
        slicer = (interval.start <= key) & (key <= interval.end)
    else:
        slicer = pd.IndexSlice[interval.start: interval.end]

    if isinstance(df, pd.DataFrame):
        sliced = df.loc[slicer, :] if axis == 0 else df.loc[:, slicer]
        if exclude:
            sliced = (
                df.loc[df.index.difference(sliced.index), :]
                if axis == 0
                else df.loc[:, df.columns.difference(sliced.columns)]
            )
    else:
        sliced = df.loc[slicer]
        if exclude:
            sliced = df.loc[df.index.difference(sliced.index)]

    return sliced


def is_frame_empty(df: Optional[Union["pd.DataFrame", "pd.Series"]]) -> bool:
    """Check if a dataframe is empty."""
    return df is None or len(df) == 0


def get_frame_window(
    df: "pd.DataFrame", column: str = None, level: int = 0
) -> Optional[TimeWindow]:
    """Get the time window of a dataframe.

    Args:
        df: dataframe
        column: If specified, the window is computed from the min and max of the column.
        level: Window is computed from the min and max of the index at the specified level.

    """
    _validate_libraries()
    if is_frame_empty(df):
        return None
    if column:
        start_time, end_time = df[column].min(), df[column].max()
    else:
        start_time, end_time = df.index.min(), df.index.max()
        if isinstance(start_time, tuple):
            start_time, end_time = start_time[level], end_time[level]

    return TimeWindow(from_timestamp(start_time.timestamp()), from_timestamp(end_time.timestamp()))


def split_frame(
    obj: Union["pd.Series", "pd.DataFrame"], n: Optional[Union[int, float, datetime]] = 0.5
) -> Union[Tuple["pd.Series", "pd.Series"], Tuple["pd.DataFrame", "pd.DataFrame"]]:
    """Split a dataframe or series into two parts.

    Args:
        obj: The object to split
        n: The index to split at. If float, it is treated as a fraction of the length of the object.
           If int, it is treated as an index.
           If datetime, it is treated as a timestamp.

    Returns:
        A tuple of two objects, the first part and the second part.
    """
    _validate_libraries()
    if isinstance(n, float):
        assert abs(n) <= 1, "Float split index should be less than equal to 1"
        n = math.floor(len(obj) * n)
    if isinstance(n, int):
        return obj.iloc[:n], obj.iloc[n:]
    else:
        return obj.loc[:n], obj.loc[n:]


def select_frame(frame: "pd.DataFrame", **conditions):
    """Select rows from a dataframe based on conditions."""
    _validate_libraries()
    for key, condition in conditions.items():
        parts = key.split("__")
        if len(parts) == 1:
            parts.append("eq")
        key, op = parts
        if isinstance(condition, Interval):
            frame = slice_frame(condition, frame, key=key, exclude=op in ("neq", "ne"))
        else:
            frame = frame[_selector(frame[key], op, condition)]
    return frame


def select_series(series: "pd.Series", **conditions):
    _validate_libraries()
    for op, condition in conditions.items():
        if isinstance(condition, Interval):
            series = series[(condition.start <= series) & (series <= condition.end)]
        else:
            series = series[_selector(series, op, condition)]
    return series


def _selector(series: "pd.Series", op, condition) -> "pd.Series":
    _validate_libraries()
    match op:
        case "eq":
            return series == condition
        case "ne" | "neq":
            return series != condition
        case "gt":
            return series > condition
        case "gte":
            return series >= condition
        case "lt":
            return series < condition
        case "lte":
            return series <= condition
        case "in" | "isin":
            return series.isin(condition)
        case "nin" | "notin":
            return ~series.isin(condition)
        case "isna" | "isnull":
            return series.isna() if condition else series.notna()



import pandas as pd

try:
    import pandas as _pd
    from vayu.pandas_utils import select_series, select_frame
    if not hasattr(_pd.DataFrame, "select"):
        _pd.DataFrame.select = select_frame
    if not hasattr(pd.Series, "select"):
        _pd.Series.select = select_series
except ImportError:
    pass



import random
import time
import threading
from abc import abstractmethod, ABC
from enum import Enum
from multiprocessing import Manager
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED, ProcessPoolExecutor
from typing import Dict, Hashable, Callable, Union

from vayu.log import L


class SafeCounter(ABC):
    @abstractmethod
    def increment(self, key):
        pass

    @abstractmethod
    def get(self, key):
        pass

    @abstractmethod
    def get_all(self):
        pass

    @staticmethod
    def make(executor_cls: type) -> "SafeCounter":
        if executor_cls == ThreadPoolExecutor:
            return ThreadSafeCounter()
        elif executor_cls == ProcessPoolExecutor:
            return ProcessSafeCounter()
        else:
            raise ValueError(f"Invalid executor type {executor_cls}")


class ThreadSafeCounter(SafeCounter):
    def __init__(self):
        self.lock = threading.Lock()
        self.d = {}

    def increment(self, key):
        with self.lock:
            self.d[key] = self.d.get(key, 0) + 1

    def get(self, key):
        with self.lock:
            return self.d.get(key, 0)

    def get_all(self):
        with self.lock:
            return self.d.copy()


class ProcessSafeCounter(SafeCounter):
    def __init__(self):
        manager = Manager()
        self.d = manager.dict()

    def increment(self, key):
        try:
            self.d[key] = self.d.get(key, 0) + 1
        except Exception:
            L.warn("Exception in ProcessSafeCounter.increment", exc_info=True)

    def get(self, key):
        return self.d.get(key, 0)

    def get_all(self):
        return self.d.copy()


class SafeDict(ABC):
    @abstractmethod
    def set(self, key, value):
        pass

    @abstractmethod
    def get(self, key):
        pass

    @abstractmethod
    def pop(self, key):
        pass

    @abstractmethod
    def get_all(self):
        pass

    @staticmethod
    def make(executor_cls: type) -> "SafeDict":
        if executor_cls == ThreadPoolExecutor:
            return ThreadSafeDict()
        elif executor_cls == ProcessPoolExecutor:
            return ProcessSafeDict()
        else:
            raise ValueError(f"Invalid executor type {executor_cls}")


class ThreadSafeDict(SafeDict):
    def __init__(self):
        self.lock = threading.Lock()
        self.d = {}

    def set(self, key, value):
        with self.lock:
            self.d[key] = value

    def get(self, key):
        with self.lock:
            return self.d.get(key)

    def pop(self, key):
        with self.lock:
            return self.d.pop(key)

    def get_all(self):
        with self.lock:
            return self.d.copy()


class ProcessSafeDict(SafeDict):
    def __init__(self):
        manager = Manager()
        self.d = manager.dict()

    def set(self, key, value):
        self.d[key] = value

    def get(self, key):
        return self.d.get(key)

    def pop(self, key):
        return self.d.pop(key)

    def get_all(self):
        return self.d.copy()


class ParallelRunner:
    """Simplified wrapper over ThreadPoolExecutor / ProcessPoolExecutor."""

    class State(Enum):
        INIT = 1
        RUNNING = 2
        DONE = 3

    def __init__(
        self,
        n_workers: int = None,
        executor_cls: type = None,
        initializer: Callable = None,
        init_args: tuple = None,
        return_when=ALL_COMPLETED,
        print_progress_every=5,
        report_exceptions=True,
    ):
        if executor_cls not in (ThreadPoolExecutor, ProcessPoolExecutor):
            raise ValueError(f"Invalid executor type {executor_cls}")

        self._n_workers = n_workers
        self._return_when = return_when
        self._tasks = dict()
        self._counter = SafeCounter.make(executor_cls)
        self._executor_cls = executor_cls if executor_cls else ThreadPoolExecutor
        self._executor = None
        self._initializer = initializer
        self._init_args = init_args or ()
        self._print_progress_every = print_progress_every
        self._state = ParallelRunner.State.INIT
        self._completion_signal: threading.Event = threading.Event()
        self._n_tasks = 0
        self._report_exceptions = report_exceptions

    def add(self, key: Hashable, func: Callable, *args, **kwargs):
        if self._state != ParallelRunner.State.INIT:
            raise RuntimeError("Cannot add tasks after calling run()")
        self._tasks[key] = (func, args, kwargs)
        self._n_tasks += 1

    def run(self) -> Dict[Hashable, object]:
        if self._state != ParallelRunner.State.INIT:
            raise RuntimeError("Cannot call run() more than once")
        self._state = ParallelRunner.State.RUNNING

        progress_thread = threading.Thread(target=self._print_progress)
        progress_thread.start()

        futures = dict()
        with self._executor_cls(
            max_workers=self._n_workers, initializer=self._initializer, initargs=self._init_args
        ) as ex:
            self._executor = ex
            for _ in range(len(self._tasks)):
                key, (func, args, kwargs) = self._tasks.popitem()
                futures[key] = ex.submit(
                    self._run, self._counter, key, self._report_exceptions, func, *args, **kwargs
                )
            try:
                wait(list(futures.values()), return_when=self._return_when)
                print("Values fetched")
            except KeyboardInterrupt:
                print("Shutting down...")
                self.shutdown()
                raise
            finally:
                self._state = ParallelRunner.State.DONE
                self._completion_signal.set()
        return {key: future.result() for key, future in futures.items()}

    def _print_progress(self):
        t = time.time()
        while self._completion_signal.wait(self._print_progress_every) is False:
            counters = self._counter.get_all()
            n_successes = counters.get("success", 0)
            n_failures = counters.get("failure", 0)
            completions = n_successes + n_failures
            perc = completions / self._n_tasks * 100
            time_taken = time.time() - t
            print(
                f"Success: {n_successes}, Failure: {n_failures}, "
                f"Completion: [{completions}/{self._n_tasks}] {perc:.2f}% "
                f"{time_taken:.1f}s ({(completions/time_taken):.2f} exec/s)",
                end="\r",
            )

    def shutdown(self, wait=True, *, cancel_futures=True):
        self._executor.shutdown(wait=wait, cancel_futures=cancel_futures)

    @staticmethod
    def _run(counter: SafeCounter, key, report_exception, f, *args, **kwargs):
        try:
            result = f(*args, **kwargs)
            counter.increment("success")
        except Exception as e:
            result = e
            counter.increment("failure")
            if report_exception:
                L.error(f"Exception in task for key {key}", exc_info=True)

        return result


if __name__ == "__main__":
    # Test ParallelRunner with ThreadPoolExecutor
    def f(x):
        time.sleep(random.randint(__ctx__["a"], __ctx__["b"]) / 10)
        return x**2

    def init():
        global __ctx__
        __ctx__ = {"a": 0, "b": 3}

    runner = ParallelRunner(
        n_workers=5,
        executor_cls=ThreadPoolExecutor,
        print_progress_every=1,
        initializer=init,
    )
    for i in range(300):
        runner.add(i, f, i)
    result = runner.run()
    print(result)



import asyncio
import time
import uuid
from contextlib import asynccontextmanager
from datetime import timedelta
from typing import Optional

try:
    from redis.asyncio.lock import Lock
    from redis.asyncio import Redis
except ImportError:
    Lock = None
    Redis = None

from vayu.aio import sleep_until_signal
from vayu.common import add_jitter


def _validate_import():
    if Lock is None or Redis is None:
        raise ImportError("redis.asyncio.lock and redis.asyncio are not available.")


class LongLivedLock(Lock):
    """A long-lived distributed lock that extends its expiration time periodically until released."""
    def __init__(self, shutdown: Optional[asyncio.Event] = None, **kwargs):
        super(LongLivedLock, self).__init__(**kwargs)
        self._shutdown = shutdown

    async def acquire(self, **kwargs):
        if not self._shutdown:
            self._shutdown = asyncio.Event()

        result = await super(LongLivedLock, self).acquire(**kwargs)
        asyncio.create_task(self._extend_lock_periodically())
        return result

    async def _extend_lock_periodically(self):
        first_loop = True
        while not self._shutdown.is_set():
            if self.local.token is None:
                return
            if first_loop:
                first_loop = False
            else:
                await self.extend(additional_time=self.timeout, replace_ttl=True)
            sleep_for = self.timeout / 4
            # Add jitter upto 20% of sleep_for.
            sleep_for = add_jitter(sleep_for, 20)
            await sleep_until_signal(duration=timedelta(seconds=sleep_for), signal=self._shutdown)

    @staticmethod
    def make_lock(
        redis: "Redis",
        name,
        timeout: timedelta = None,
        sleep: timedelta = None,
        blocking_timeout: timedelta = None,
        shutdown: Optional[asyncio.Event] = None,
    ) -> "LongLivedLock":
        return LongLivedLock(
            # TODO: See if we can get away with using a single redis connection.
            redis=redis,
            name=f"lock.{name}",
            timeout=timeout.total_seconds() if timeout else 60,
            sleep=sleep.total_seconds() if sleep else 5,
            blocking_timeout=blocking_timeout.total_seconds() if blocking_timeout else None,
            shutdown=shutdown,
        )


class RedisSemaphore:
    def __init__(self, redis_client: Redis, key: str, limit: int, ttl: timedelta = None, retry_every: timedelta = None, keep_renewing: bool = False):
        """A distributed semaphore using redis sorted sets.

        Caveats:

        Args:
            redis_client:
            key: Key to use for the semaphore.
            limit: Maximum number of concurrent users.
            ttl: Time to live for each user's lock.
            retry_every:
            keep_renewing:
        """
        self._client = redis_client
        self._key = key
        self._limit = limit
        self._ttl = ttl or timedelta(seconds=60)
        self._retry_every = retry_every or timedelta(seconds=5)
        self._keep_renewing = keep_renewing

        # Lua script to add to sorted set if it is not full.
        lua_script = """
        local key = KEYS[1]
        local member = ARGV[1]
        local score = ARGV[2]
        local min_score = ARGV[3]
        local n = tonumber(ARGV[4])
        redis.call('ZREMRANGEBYSCORE', key, '-inf', min_score)
        if redis.call('ZCARD', key) < n then
            return redis.call('ZADD', key, score, member)
        else
            return nil
        end
        """

        self._lua_script = self._client.register_script(lua_script)

    async def _add_to_sorted_set(self, member: str, score: int, min_score: int):
        result = await self._lua_script(keys=[self._key], args=[member, str(score), str(min_score), self._limit])
        return result is not None

    async def acquire(self, identifier: str, shutdown: Optional[asyncio.Event] = None):
        while True:
            score = round(time.time())
            min_score = score - self._ttl.total_seconds()
            if await self._add_to_sorted_set(member=identifier, score=int(score), min_score=int(min_score)):
                return True
            # Retry after some time with some jitter.
            if await sleep_until_signal(duration=add_jitter(self._retry_every), signal=shutdown or asyncio.Event()):
                # Shutdown signal received.
                return False

    async def release(self, identifier: str):
        await self._client.zrem(self._key, identifier)

    @asynccontextmanager
    async def lock(self, identifier: str = None, shutdown: Optional[asyncio.Event] = None):
        identifier = identifier or uuid.uuid4().bytes
        acquired = False
        sig_renew = asyncio.Event()
        task_renew = None

        try:
            acquired = await self.acquire(identifier, shutdown=shutdown)
            if acquired and self._keep_renewing:
                task_renew = asyncio.create_task(self._renew_periodically(identifier, shutdown=sig_renew))
            yield
        finally:
            if acquired:
                await self.release(identifier)
            if task_renew:
                sig_renew.set()
                await task_renew

    async def _renew_periodically(self, identifier: str, shutdown: asyncio.Event):
        while not shutdown.is_set():
            # Renew lock every quarter of ttl only if it is still in the set.
            await self._client.zadd(self._key, {identifier: int(round(time.time()))}, xx=True)
            await sleep_until_signal(duration=add_jitter(self._ttl/4), signal=shutdown)



import os
import subprocess
import argparse
import tempfile
from pathlib import Path


def clone_repo(git_url, clone_path):
    """Clone a git repository to the specified path."""
    subprocess.run(["git", "clone", git_url, clone_path], check=True)


def concatenate_files(repo_path, output_file, extensions, repo_title):
    """Concatenate files with specified extensions into a single output file."""
    with open(output_file, "w", encoding="utf-8") as outfile:
        for root, _, files in os.walk(repo_path):
            for file in files:
                if any(file.endswith(ext) for ext in extensions):
                    file_path = os.path.join(root, file)
                    try:
                        with open(file_path, "r", encoding="utf-8") as infile:
                            contents = infile.read()
                            relative_path = file_path.replace(repo_path, repo_title)
                            outfile.write(f"--- {relative_path} ---\n{contents}\n\n")
                    except UnicodeDecodeError:
                        print(f"Skipping non-text file: {file_path}")


def parse_github_url(git_url):
    """Extract username and repository name from a GitHub URL."""
    parts = git_url.split('/')
    repo_name = parts[-1].split('.')[0] if parts[-1].endswith('.git') else parts[-1]
    github_username = parts[-2].split(':')[-1]
    return github_username, repo_name


def main():
    parser = argparse.ArgumentParser(description='Clone a GitHub repo and concatenate files into a single file.')
    parser.add_argument('--git_url', help='GitHub repository URL', default=None)
    parser.add_argument('--dir', help='Directory path to process', default=None)
    parser.add_argument('--extensions', nargs='+', default=['.py'], 
                        help='List of file extensions to include (default: .py)')
    parser.add_argument('--output_dir', help='Output directory', default=None)

    args = parser.parse_args()

    if not (args.git_url or args.dir):
        print("Error: Neither git_url nor dir is provided")
        return

    # Process extensions to ensure they start with a dot
    extensions = [ext if ext.startswith('.') else f'.{ext}' for ext in args.extensions]
    output_dir = args.output_dir if args.output_dir else os.getcwd()
    
    if args.git_url:
        github_username, repo_name = parse_github_url(args.git_url)
        repo_title = f"{github_username}/{repo_name}"
        output_filename = f"{github_username}__{repo_name}.txt"
        
        with tempfile.TemporaryDirectory(dir="/tmp") as directory:
            clone_repo(args.git_url, directory)
            output_path = os.path.join(output_dir, output_filename)
            concatenate_files(directory, output_path, extensions, repo_title)
    else:
        directory = args.dir
        repo_title = args.dir
        output_filename = f"{Path(args.dir).name}.txt"
        output_path = os.path.join(output_dir, output_filename)
        concatenate_files(directory, output_path, extensions, repo_title)

    print(f"Output saved to {output_path}")


if __name__ == "__main__":
    main()



import datetime
import time

from dataclasses import dataclass
import datetime as dt
from typing import Union
from functools import wraps
from zoneinfo import ZoneInfo

from vayu.common import Interval, TZ


def epoch_time() -> dt.datetime:
    return from_timestamp(0)


def time_now(with_ms=False, local=True) -> dt.datetime:
    t = dt.datetime.now(tz=TZ if local else datetime.UTC)
    if not with_ms:
        t = t.replace(microsecond=0)

    return t


def min_time(date: dt.date, tz=None) -> dt.datetime:
    return dt.datetime.combine(date, dt.time.min, tzinfo=tz or dt.UTC)


def max_time(date: dt.date, tz=None) -> dt.datetime:
    return dt.datetime.combine(date, dt.time.max, tzinfo=tz or dt.UTC)


def ts_ms(t: datetime.datetime) -> int:
    return int(t.timestamp() * 1000)


def utc_datetime(
    year: int,
    month: int,
    day: int,
    hour: int = 0,
    minute: int = 0,
    second: int = 0,
    microsecond: int = 0,
) -> dt.datetime:
    return dt.datetime(
        year=year,
        month=month,
        day=day,
        hour=hour,
        minute=minute,
        second=second,
        microsecond=microsecond,
        tzinfo=datetime.UTC,
    )


def local_datetime(
    year: int,
    month: int,
    day: int,
    hour: int = 0,
    minute: int = 0,
    second: int = 0,
    microsecond: int = 0,
    tz=TZ,
) -> datetime.datetime:
    t = datetime.datetime(
        year=year,
        month=month,
        day=day,
        hour=hour,
        minute=minute,
        second=second,
        microsecond=microsecond,
    )
    return t.replace(tzinfo=tz)


def from_timestamp(ts: float, tz=TZ) -> dt.datetime:
    if ts > 1e12:
        # ts is in ms.
        ts = ts / 1000
    return dt.datetime.fromtimestamp(ts, tz=tz or dt.UTC)


def from_z_string(t: str) -> dt.datetime:
    return dt.datetime.fromisoformat(t.upper().replace("Z", "+00:00"))


@dataclass
class TimeWindow(Interval):
    start: Union[dt.datetime, dt.date, None] = None
    end: Union[dt.datetime, dt.date, None] = None

    @property
    def duration(self) -> dt.timedelta:
        return self.end - self.start

    @property
    def start_ms(self) -> int:
        return int(self.start.timestamp() * 1000)

    @property
    def end_ms(self) -> int:
        return int(self.end.timestamp() * 1000)

    @staticmethod
    def ahead(
        *,
        t: dt.datetime = None,
        duration: dt.timedelta = None,
        days=0,
        seconds=0,
        microseconds=0,
        milliseconds=0,
        minutes=0,
        hours=0,
        weeks=0,
    ) -> "TimeWindow":
        if not duration:
            duration = dt.timedelta(
                days=days,
                seconds=seconds,
                microseconds=microseconds,
                milliseconds=milliseconds,
                minutes=minutes,
                hours=hours,
                weeks=weeks,
            )
        else:
            assert days == seconds == milliseconds == milliseconds == minutes == hours == weeks == 0
        t = t or time_now()
        return TimeWindow(t, t + duration)

    @staticmethod
    def behind(
        *,
        t: dt.datetime = None,
        duration: dt.timedelta = None,
        days=0,
        seconds=0,
        microseconds=0,
        milliseconds=0,
        minutes=0,
        hours=0,
        weeks=0,
    ) -> "TimeWindow":
        if not duration:
            duration = dt.timedelta(
                days=days,
                seconds=seconds,
                microseconds=microseconds,
                milliseconds=milliseconds,
                minutes=minutes,
                hours=hours,
                weeks=weeks,
            )
        else:
            assert days == seconds == milliseconds == milliseconds == minutes == hours == weeks == 0

        t = t or time_now()
        return TimeWindow(t - duration, t)

    @staticmethod
    def around(
        *,
        t: dt.datetime = None,
        duration: dt.timedelta = None,
        days=0,
        seconds=0,
        microseconds=0,
        milliseconds=0,
        minutes=0,
        hours=0,
        weeks=0,
    ):
        if not duration:
            duration = dt.timedelta(
                days=days,
                seconds=seconds,
                microseconds=microseconds,
                milliseconds=milliseconds,
                minutes=minutes,
                hours=hours,
                weeks=weeks,
            )
        else:
            assert days == seconds == milliseconds == milliseconds == minutes == hours == weeks == 0

        t = t or time_now()
        return TimeWindow(t - duration, t + duration)

    @staticmethod
    def from_date(year: int, month: int, day: int, tz=None) -> "TimeWindow":
        tz = tz or ZoneInfo("UTC")
        date = dt.date(year, month, day)
        return TimeWindow(start=min_time(date, tz=tz), end=max_time(date, tz=tz))

    @staticmethod
    def from_timestamp(start_ts: [int, float], end_ts: [int, float], tz=None) -> "TimeWindow":
        start = from_timestamp(start_ts, tz=tz)
        end = from_timestamp(end_ts, tz=tz)
        return TimeWindow(start, end)

    def __str__(self):
        if isinstance(self.start, dt.date):
            return f"[{self.start.isoformat()} - {self.end.isoformat()}]"
        return f"[{self.start.replace(microsecond=0).isoformat()} - {self.end.replace(microsecond=0).isoformat()}]"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((self.end, self.start))

    def _validate(self):
        if self.start == self.end is None:
            is_date = False
        else:
            # Note: we cannot do isinstance(t, dt.date) as it'd be True even if t was a datetime.
            is_date = not (isinstance(self.start, dt.datetime) or isinstance(self.end, dt.datetime))
        if self.start and self.end:
            assert type(self.start) is type(self.end), "Both start and end should be of same type"

        if self.start is None:
            self.start = epoch_time().date() if is_date else epoch_time()
        if self.end is None:
            self.end = time_now().date() if is_date else time_now()
        if not is_date and bool(self.start.tzinfo) != bool(self.end.tzinfo):
            raise ValueError("end and start should either be tz aware or unaware.")

        assert self.start <= self.end, "start time should be less than equal to end time"

    def __post_init__(self):
        self._validate()


def timeit(f):
    @wraps(f)
    def timed(*args, **kw):

        t = -time.time()
        result = f(*args, **kw)
        t += time.time()

        print(f"`{f.__name__}` took: {round(t, 3)}s")
        return result

    return timed


def to_human_readable_time(t: Union[dt.timedelta, float]):
    if isinstance(t, dt.timedelta):
        t = int(t.total_seconds())
    if t < 0:
        return "invalid-time"

    h, remainder = divmod(t, 3600)
    m, s = divmod(remainder, 60)

    parts = []
    if h > 0:
        parts.append(f"{h}h")
    if m > 0:
        parts.append(f"{m}m")
    if s > 0:
        parts.append(f"{s}s")

    return "".join(parts) if parts else "0s"



import asyncio
import secrets
from abc import ABC, abstractmethod
from datetime import timedelta
from functools import wraps
from hashlib import sha256
from pathlib import Path
from time import time
from typing import Callable, Any, Union, Dict, Tuple, Optional
import pickle
from weakref import WeakValueDictionary

from aiofiles import os as async_os
from aiofiles import open as async_open

try:
    import orjson
    _use_orjson = True
except ImportError:
    import json
    _use_orjson = False

from vayu.log import L
from vayu.time import time_now, from_timestamp


class _MethodDecoratorAdaptor(object):
    def __init__(self, decorator, func):
        self.decorator = decorator
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.decorator(self.func)(*args, **kwargs)

    def __get__(self, instance, owner):
        return self.decorator(self.func.__get__(instance, owner))


def auto_adapt_to_methods(decorator):
    """Allows you to use the same decorator on methods and functions,
    hiding the self argument from the decorator."""

    def adapt(func):
        return _MethodDecoratorAdaptor(decorator, func)

    return adapt


class Serializer(ABC):
    @abstractmethod
    def serialize(self, obj: Any) -> bytes:
        pass

    @abstractmethod
    def deserialize(self, bts: bytes) -> Any:
        pass


class Pickler(Serializer):
    def serialize(self, obj: Any) -> bytes:
        return pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)

    def deserialize(self, bts: bytes) -> Any:
        return pickle.loads(bts)


class Jsoner(Serializer):
    def serialize(self, obj: Any) -> bytes:
        if _use_orjson:
            return orjson.dumps(obj)
        else:
            return json.dumps(obj).encode()

    def deserialize(self, bts: bytes) -> Any:
        if _use_orjson:
            return orjson.loads(bts)
        else:
            return json.loads(bts.decode())


class CacheMissError(Exception):
    pass


class Cache(ABC):
    def __init__(self, serializer: Serializer):
        self._serializer = serializer
        self._locks = dict()

    def cached(
            self,
            ttl: Union[timedelta, float],
            prefix: str = None,
            key: Union[str, Callable] = None,
            serializer: Serializer = None,
            log: bool = True,
    ):
        """Used as decorator to cache function calls.

        Args:
            ttl: cache expiry ttl
            prefix: If set, gets prepended to the rest of the cache key, otherwise uses function,
                        class, and module name to create prefix.
            key:
               - None: Use SHA256 digest of *args & **kwargs to construct cache key
               - str: Use the specified string and ignore function arguments
               - Callable: Calls it with *args * **kwargs which returns a str key.
            serializer: If not specified, default pickler serializer is used
            log: If True, logs reports such as cache hit, miss etc.

        """
        serializer = serializer or self._serializer
        if ttl is None:
            raise ValueError("ttl must be specified for cached decorator")

        if ttl is not None:
            if not isinstance(ttl, timedelta):
                ttl = timedelta(seconds=ttl)
            if ttl.total_seconds() <= 0:
                raise ValueError("ttl must be positive")

        @auto_adapt_to_methods
        def decorator(func):
            f_name = func.__name__
            if not asyncio.iscoroutinefunction(func):
                raise NotImplementedError("@cached can be used only with coroutines.")

            @wraps(func)
            async def async_inner(*args, **kwargs):
                t = time()
                k = self._key(func, prefix=prefix, key=key, args=args, kwargs=kwargs)
                t_del = time() - t
                if 0.001 < t_del and log:
                    L.warn(f"Key construction for cached fn `{f_name}` took too long: {t_del:.6f}s")

                try:
                    got = await self._read(k)
                    return serializer.deserialize(got)
                except CacheMissError:
                    pass

                async with self._locks.setdefault(k, asyncio.Lock()):
                    try:
                        got = await self._read(k)
                        return serializer.deserialize(got)
                    except CacheMissError:
                        pass

                t = time()
                result = await func(*args, **kwargs)
                t_del = time() - t
                if log:
                    L.info(f"@cached({f_name}) - CACHE MISS | took: {t_del:.6}s")
                await self._write(k, serializer.serialize(result), ttl)
                return result

            @wraps(func)
            async def evict(*args, **kwargs):
                k = self._key(func, prefix=prefix, key=key, args=args, kwargs=kwargs)
                await self._delete(k)

            async_inner.evict = evict

            return async_inner

        return decorator

    async def read(self, key: str) -> Any:
        got = await self._read(key)
        return self._serializer.deserialize(got)

    async def write(self, key: str, value: Any, ttl: timedelta):
        return await self._write(key, self._serializer.serialize(value), ttl)

    @abstractmethod
    async def _read(self, key: str) -> Any:
        raise NotImplementedError

    @abstractmethod
    async def _write(self, key: str, value: Any, ttl: timedelta):
        raise NotImplementedError

    @abstractmethod
    async def _delete(self, key: str):
        raise NotImplementedError

    @staticmethod
    def _key(
            fn: Callable,
            prefix: str = None,
            key: Union[str, Callable] = None,
            args: Tuple = None,
            kwargs: Dict = None,
    ) -> str:
        is_method = hasattr(fn, "__self__")

        if isinstance(key, str):
            return key

        parts = []
        if prefix:
            parts.append(prefix)
        else:
            parts.append(str(fn.__module__))
            if is_method:
                parts.append(str(fn.__self__.__class__.__name__))
            parts.append(fn.__name__)
        parts = [".".join(parts)]
        if callable(key):
            parts.append(key(*args, **kwargs))
        else:
            # Sort kwargs to ensure deterministic key.
            kwargs_key = [f"{k}={v}" for k, v in sorted(kwargs.items(), key=lambda x: x[0])]
            arg_key = ",".join([str(a) for a in args] + kwargs_key)
            parts.append(sha256(arg_key.encode()).hexdigest())
        return ":".join(parts)


class FileCache(Cache):
    def __init__(self, path: str, serializer: Serializer = None):
        super().__init__(serializer or Pickler())
        self._path = Path(path)
        self._ttl_bytes_length = 8
        self._path.mkdir(parents=True, exist_ok=True)

    def _filename(self, key: str) -> Path:
        return self._path / key

    async def _write(self, key: str, value: bytes, ttl: Optional[timedelta]):
        expiry_time = time_now(with_ms=True) + ttl
        ttl_val = int(expiry_time.timestamp() * 1e6)
        ttl_bytes = ttl_val.to_bytes(self._ttl_bytes_length, "big")
        bts = ttl_bytes + value

        final_path = self._filename(key)
        tmp_name = f".{final_path.name}.tmp.{secrets.token_hex(4)}"
        tmp_path = self._path / tmp_name

        async with async_open(tmp_path, "wb") as f:
            await f.write(bts)

        # Atomic replace (POSIX-atomic on same filesystem)
        await async_os.replace(tmp_path, final_path)

    async def _read(self, key: str) -> bytes:
        filename = self._filename(key)
        try:
            async with async_open(filename, "rb") as f:
                ttl_bytes = await f.read(self._ttl_bytes_length)
                if len(ttl_bytes) < self._ttl_bytes_length:
                    # corrupt file
                    raise CacheMissError()
                ttl_val = int.from_bytes(ttl_bytes, "big")
                expiry = from_timestamp(ttl_val / 1e6)  # reverse of write
                if expiry < time_now(with_ms=True):
                    raise CacheMissError()
                else:
                    bts = await f.read()
        except (FileNotFoundError, Exception):
            raise CacheMissError()

        return bts

    async def _delete(self, key: str):
        filename = self._filename(key)
        try:
            await async_os.remove(filename)
        except FileNotFoundError:
            pass


class MemoryCache(Cache):
    class NoOpSerializer(Serializer):
        def serialize(self, obj: Any) -> Any:
            return obj

        def deserialize(self, bts: Any) -> Any:
            return bts

    def __init__(self):
        super().__init__(self.NoOpSerializer())
        self._cache = {}

    async def _write(self, key: str, value: Any, ttl: timedelta):
        self._cache[key] = (time_now(with_ms=True) + ttl, value)

    async def _read(self, key: str) -> Optional[Any]:
        try:
            expiry_time, obj = self._cache[key]
        except KeyError:
            raise CacheMissError()
        if expiry_time < time_now(with_ms=True):
            await self._delete(key)
            raise CacheMissError()
        return obj

    async def _delete(self, key: str):
        try:
            del self._cache[key]
        except KeyError:
            L.info(f"Key {key} was not found in cache")
            pass


mem_cache = MemoryCache()
mem_cached = mem_cache.cached



import numpy as np
import pandas as pd


def quantile(q) -> callable:
    """Return a function that computes the quantile of a series.

    Notes:
        - Helpful as an aggregation function.
        - About 2.5x faster than pandas.Series.quantile.
    """

    def inner(a):
        if isinstance(a, pd.Series):
            a = a.values
        if len(a) == 0:
            return np.nan
        return np.quantile(a, q)

    inner.__name__ = f"q{q}"

    return inner



import asyncio
from dataclasses import dataclass
from datetime import timedelta
from numbers import Number
from signal import SIGTERM, SIGINT
from typing import Union, Coroutine, Optional, List, Dict, Any

import aiohttp
from aiohttp import TCPConnector


async def get_first_and_cancel_rest(*tasks: Union[asyncio.Task, Coroutine]) -> asyncio.Task:
    done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
    for task in pending:
        task.cancel()

    return done.pop()


async def sleep_until_signal(duration: Union[timedelta, Number], signal: asyncio.Event) -> bool:
    """Sleep until duration unless signal fires. Returns True if signal fired first."""
    if isinstance(duration, (int, float)):
        duration = timedelta(seconds=duration)

    sleep_task = asyncio.create_task(asyncio.sleep(duration.total_seconds()))
    signal_task = asyncio.create_task(signal.wait())
    done = await get_first_and_cancel_rest(sleep_task, signal_task)
    return done is signal_task


def attach_shutdown_signals(shutdown: asyncio.Event, signals: Optional[List] = None):
    signals = signals or [SIGINT, SIGTERM]
    loop = asyncio.get_running_loop()

    def set_signal():
        shutdown.set()

    for signal in signals:
        loop.add_signal_handler(signal, set_signal)


@dataclass(frozen=True)
class RateLimit:
    limit: int
    period: timedelta


async def grab_all_urls(
        key_url_map: Dict[Any, str],
        concurrency=100,
        timeout=15,
        print_progress=True,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[aiohttp.BasicAuth] = None,
        rate_limit: Optional[RateLimit] = None
) -> Dict[Any, Any]:
    results = {}
    tasks = []

    semaphore = asyncio.Semaphore(concurrency)

    async def fetch_url(session, url, key, results: Dict[Any, Any], timeout):
        async with semaphore:
            try:
                async with session.get(url, timeout=timeout) as response:
                    results[key] = (response.status, await response.read())
            except Exception as e:
                results[key] = None, str(e)

    async with aiohttp.ClientSession(connector=TCPConnector(limit=concurrency), headers=headers, auth=auth) as session:
        for key, url in key_url_map.items():
            task = fetch_url(session, url, key, results, timeout)
            tasks.append(task)

        for i, task in enumerate(asyncio.as_completed(tasks)):
            await task
            if print_progress:
                print(f'Progress: {i+1}/{len(tasks)}', end="\r")

    return results



import logging
import os
import sys
import datetime
from pytz import timezone

LOGGING_TZ = timezone(os.getenv("TZ", "UTC"))


def _format_time(self, record, _=None):
    return (
        datetime.datetime.fromtimestamp(record.created, datetime.timezone.utc)
        .replace(microsecond=0)
        .astimezone(LOGGING_TZ)
        .isoformat(sep="T")
    )


formatter = logging.Formatter("[%(asctime)s] %(levelname)s> %(message)s")
logging.Formatter.formatTime = _format_time

_asyncio_logger = logging.getLogger("asyncio")
L = logging.getLogger()

_asyncio_logger.setLevel(logging.WARNING)
L.setLevel(os.getenv("LOG_LEVEL", "INFO").upper())

_handler = logging.StreamHandler(sys.stdout)
_handler.setLevel(logging.INFO)
_handler.setFormatter(formatter)
L.addHandler(_handler)
_asyncio_logger.addHandler(_handler)

L.i = L.info
L.d = L.debug
L.e = L.error
L.w = L.warning
L.c = L.critical

__all___ = [L]
